#!/bin/sh
#
# Copyright 2022 ProAtria team
#
# Script for initializing the SFTPPlus installation on Linux / Unix.
# Uses the default /bin/sh shell. Tested with the OS'es supported by SFTPPlus.

if ! command -v dirname >/dev/null 2>&1; then
    printf "Missing \"dirname\" command.\n"
    exit 254
fi

# POSIX shell assurances.
set -o nounset
set -o errexit

bin_relative_path="$(dirname "$0")"

# Using pwd in a sub-shell, get the absolute path to the sub-dir of this script.
bin_path=$(cd "$bin_relative_path" && pwd)

if [ -z "$bin_path" ]; then
    printf "Couldn't determine absolute path for \"%s\".\n" "$0"
    exit 253
fi

export INSTALL_DIR="${bin_path%/bin}"

if [ "$(uname)" = "AIX" ]; then
    # Avoid the error "@: parameter not set" when there is no parameter,
    # as the behaviour of the default /bin/sh is different on AIX.
    set +o nounset
fi

set_custom_service_name() {
    # Chop "service-name=" from the parameter and set the var needed later.
    export SFTPPLUS_SERVICE="${1#"service-name="}"

    if [ -z "$SFTPPLUS_SERVICE" ]; then
        (>&2 printf "Empty service names are not allowed.\n")
        exit 106
    fi

    printf "Custom SFTPPlus service name provided: %s\n" "$SFTPPLUS_SERVICE"
}

# Show this help message and exit.
# Whenever editing this, check alignment of opts / associated text and length.
command_show_help() {
    printf "\nUsage:\n"
    printf "  %s [OPTIONS] PATH_TO_TAR_GZ\n" "$0"
    printf "\nOptions:\n"
    printf "  -h, --help\t\t\tShow this help message and exit\n"
    printf "      --service-name=NAME"
    printf "\tSet a custom NAME for the SFTPPlus service\n\n"
    printf "With the --service-name option, you can define a custom name for\n"
    printf "the SFTPPlus system service, for example \"sftpplus-production\".\n"
    printf "The default option is \"sftpplus-mft\".\n"
    printf "\nPATH_TO_TAR_GZ is the path to a SFTPPlus installation package,\n"
    printf "for example \"sftpplus-linux-x64-trial.tar.gz\".\n"
    exit 0
}

# Get command line parameters and act accordingly.
while getopts "h-:" opt; do
    case $opt in
        h)
            command_show_help
            ;;
        -)
            # getopts doesn't support long opts such as --help.
            case "$OPTARG" in
                help)
                    command_show_help
                    ;;
                service-name=*)
                    set_custom_service_name "$OPTARG"
                    ;;
                *)
                    printf "Invalid option: --%s\n" "$OPTARG"
                    exit 1
                    ;;
            esac
            ;;
        \?)
            # Unknown option.
            exit 1
            ;;
    esac
done

exec /bin/sh "$INSTALL_DIR"/bin/setup_helper.sh install "$@"
