#!/bin/sh
#
# Script helper for initializing / updating / rolling back / uninstalling
# SFTPPlus on Linux, macOS, AIX.
# To be used only from the corresponding shell scripts.

# POSIX shell assurances.
set -o nounset
set -o errexit

LATEST_VER_LNK="https://download.sftpplus.com/sftpplus/LATEST"
LNX_DOCS="https://sftpplus.com/doc-4-linux"
LNX_ADV_DOCS="https://sftpplus.com/doc-4-linux-advance"
LNX_PORT_DOCS="https://sftpplus.com/doc-4-linux-port"
UPGRADE_DOCS="https://sftpplus.com/doc-4-upgrade"
UNINST_DOCS="https://sftpplus.com/doc-4-uninstall"

CONF_FILE="configuration/server.ini"
INSTALL_CONF="${INSTALL_DIR:-}/$CONF_FILE"
# FIXME:5956:
# Look at handling the INSTALL_INFO file via Python CLI.
INSTALL_INFO="${INSTALL_DIR:-}/configuration/INSTALL_INFO"
ADMIN_CMD="${INSTALL_DIR:-}/bin/admin-commands.sh"
PID_FILE="${INSTALL_DIR:-}/run/server.pid"
POST_UPDATE_SCRIPT="${INSTALL_DIR:-}/bin/post-update.sh"
DATE_STRING="$(date '+%Y%m%d-%H%M%S')"
SUBDIRS_TO_REMOVE="bin doc lib run"
SUBDIRS_TO_CHOWN="configuration log run"
# Spaces are not allowed in the backup path,
# which must be on a filesystem with support for file capabilities.
# This path is hardcoded in bin/rollback.sh and also appears in documentation.
BKP_COMMON_DIR="/opt/sftpplus_backups"
BKP_STRING="AUTO_BACKUP"
BKP_SUFFIX="${BKP_STRING}_${DATE_STRING}"
DEV_STRING="dev"
MIN_GLIBC2_VER="5"
PORT_LM="10020"
PORT_HTTPS="10443"

# 'account' parameter to be set in configuration/server.ini, section [server].
SFTPPLUS_USER=""
# This has a default value, as it's not asked for.
# Might also be missing from INSTALL_INFO.
# But it can be set as a command line option when installing (see install.sh).
SFTPPLUS_SERVICE="${SFTPPLUS_SERVICE:-sftpplus-mft}"
DROP_OPTION=""

# Common init-specific dirs holding our service / unit / init files.
SYSTEMD_INIT_DIR="/etc/systemd/system"
OPENRC_INIT_DIR="/etc/init.d"
LAUNCHD_INIT_DIR="/Library/LaunchDaemons"
AIX_INIT_DIR="/etc/rc.d/init.d"

# English only. On macOS, tr complains of illegal bytes w/ default UTF-8 locale.
export LC_ALL=C
LOCAL_IP="ip_not_yet_determined"
PROCESS_STOP_COUNTER="0"
# Export it to have it in the post-update.sh script as well.
export SELINUX_ENFORCED="no"

error_out_with_generic_message() {
    if [ $? != 0 ]; then
        printf "\nThe %s action failed. Try correcting the error above.\n" \
            "$ACTION"
        printf "Alternatively, contact us at support@proatria.com for help.\n"
    fi
}

# export is used here to assign a variable to another variable.
read_with_default() {
    read -r read_value
    if [ -z "$read_value" ]; then
        export "$1"="$2"
    else
        export "$1"="$read_value"
    fi
}

yes_or_default_not() {
    printf "\n%s (Yes/[No]): " "$1"
    read -r read_value

    case $read_value in
        "yes"|"y"|"YES"|"Y"|"Yes")
            YES_OR_NOT="yes"
            ;;
        "no"|"n"|"NO"|"N"|"No"|"")
            YES_OR_NOT="no"
            ;;
        *)
            (>&2 printf "Ambiguous answer: %s\n" "$read_value")
            yes_or_default_not "$1"
            ;;
    esac
}


# Check Python binary is present and working, so that initialization works.
# Otherwise, install data would be collected from the user needlessly.
# While at it, get current version.
check_admin_cmd() {
    if [ ! -x "$ADMIN_CMD" ]; then
        (>&2 printf "\"%s\" not found or not executable.\n" "$ADMIN_CMD")
        (>&2 printf "Make sure SFTPPlus files were unpacked successfully.\n")
        exit 69
    else
        # Firstly, check without redirecting stderr, to have errors visible.
        "$ADMIN_CMD" generate-uuid >/dev/null
        # This option outputs to stderr.
        VERSION=$("$ADMIN_CMD" --version 2>&1)
        # Export it to have it in the post-update.sh script as well.
        export VERSION
    fi
}

# FIXME:4550:
# The logic from here should be moved into the Python CLI handling.
initialize_config() {
    printf "SFTPPlus administrator name to use (admin): "
    read_with_default ADMIN_USER admin
    # Randomness can be a problem when virtualized, use openssl if possible.
    if which openssl >/dev/null 2>/dev/null; then
        RANDOM_PASS=$(openssl rand -base64 12)
    else
        RANDOM_PASS=$(tr -cd A-Za-z0-9 </dev/urandom | head -c 12)
    fi
    printf "SFTPPlus administrator's password (%s): " "$RANDOM_PASS"
    read_with_default ADMIN_PASS "$RANDOM_PASS"

    printf "\nInitializing SFTPPlus configuration...\n"
    # Do not quote $DROP_OPTION to have this working when it's empty.
    # shellcheck disable=SC2086
    "$ADMIN_CMD" initialize \
        --init-admin "$ADMIN_USER" \
        --init-password "$ADMIN_PASS" \
        --service-os-user "$SFTPPLUS_USER" \
        $DROP_OPTION
}

macos_user_setup() {
    # Create the group dedicated to the service account.
    dscl . create /Groups/"$SFTPPLUS_USER"
    # Assign an unique ID to the group.
    dscl . create /Groups/"$SFTPPLUS_USER" PrimaryGroupID "$1"
    # Disable group password.
    dscl . create /Groups/"$SFTPPLUS_USER" Password '*'
    # Create a user for the service account.
    dscl . create /Users/"$SFTPPLUS_USER"
    # This is similar to the GECOS field for the user's full name.
    dscl . create /Users/"$SFTPPLUS_USER" RealName "$GECOS_DATA"
    # Assign a unique ID to the new user.
    dscl . create /Users/"$SFTPPLUS_USER" UniqueID "$1"
    # Assign this account to the dedicated group.
    dscl . create /Users/"$SFTPPLUS_USER" PrimaryGroupID "$1"
    # Disable shell access.
    dscl . create /Users/"$SFTPPLUS_USER" UserShell /usr/bin/false
    # Make sure it has a default empty home folder.
    dscl . create /Users/"$SFTPPLUS_USER" NFSHomeDirectory /var/empty
    # Disable password to block any authentication request.
    dscl . create /Users/"$SFTPPLUS_USER" Password '*'
    # Initialize blank password and authentication rules.
    dscl . delete /Users/"$SFTPPLUS_USER" PasswordPolicyOption
    dscl . delete /Users/"$SFTPPLUS_USER" AuthenticationAuthority
}

add_os_user() {
    GECOS_DATA="Added by SFTPPlus $VERSION on $DATE_STRING"
    case "$OS" in
        "AIX")
            mkgroup "$SFTPPLUS_USER"
            mkuser "$SFTPPLUS_USER"
            usermod -g "$SFTPPLUS_USER" -d "$INSTALL_DIR" -s /bin/sh \
                -c "$GECOS_DATA" "$SFTPPLUS_USER"
            ;;
        "Darwin")
            printf "Unique ID for the %s OS user/group (299): " "$SFTPPLUS_USER"
            read_with_default MACOS_ID 299
            macos_user_setup "$MACOS_ID"
            ;;
        "Linux")
            if [ "$LNX_DISTRO" = "alpine" ]; then
                # Can't count on the shadow package being installed on Alpine.
                addgroup "$SFTPPLUS_USER"
                adduser -G "$SFTPPLUS_USER" -g "$GECOS_DATA" \
                    -s /bin/sh -h "$INSTALL_DIR" -H -D "$SFTPPLUS_USER"
            else
                # $INSTALL_DIR is preferred as it's handy when containerised.
                SFTPPLUS_HOME="$INSTALL_DIR"
                # SELinux prevents services to run scripts from a user's home.
                if [ "$SELINUX_ENFORCED" = "yes" ]; then
                    mkdir /var/lib/"$SFTPPLUS_USER"
                    SFTPPLUS_HOME=/var/lib/"$SFTPPLUS_USER"
                fi
                groupadd "$SFTPPLUS_USER"
                useradd -g "$SFTPPLUS_USER" -c "$GECOS_DATA" \
                    -s /bin/sh -d "$SFTPPLUS_HOME" -M "$SFTPPLUS_USER"
            fi
            ;;
    esac
}

# FIXME:4550:
# The logic from here should be moved into the Python CLI handling.
get_daemon_and_sftpplus_users() {
    yes_or_default_not "Do you need to authenticate OS users?"
    if [ "$YES_OR_NOT" = "yes" ]; then
        if [ "$OS" != "Linux" ]; then
            # Dropping privileges on AIX and macOS is not supported currently.
            # Also an issue when trying to identify SFTPPlus processes by user.
            (>&2 printf "\nSetting this up automatically ")
            (>&2 printf "is not supported on the current OS.\n")
            printf "Contact us at support@proatria.com for help on the issue.\n"
            exit 75
        fi
        DAEMON_USER="root"
        DROP_OPTION="--drop-user"
        printf "SFTPPlus service is going to start as root.\n"
        printf "Superuser privileges are dropped as soon as possible.\n"
        printf "New operating system account to drop privileges to (sftpplus): "
        read_with_default SFTPPLUS_USER sftpplus
        if [ "$SFTPPLUS_USER" = "root" ]; then
            printf "Running SFTPPlus as root at all times is not supported\n."
            exit 77
        fi
    else
        printf "New operating system account to run SFTPPlus (sftpplus): "
        read_with_default DAEMON_USER sftpplus
        SFTPPLUS_USER="$DAEMON_USER"
    fi

    {
        printf "DAEMON_USER=%s\n" "$DAEMON_USER"
        printf "SFTPPLUS_SERVICE=%s\n" "$SFTPPLUS_SERVICE"
        printf "SFTPPLUS_USER=%s\n" "$SFTPPLUS_USER"
        printf "SFTPPLUS_GROUP=%s\n" "$SFTPPLUS_USER"
    } > "$INSTALL_INFO"
}

# Extended ACLs can make sensitive files readable by an unwanted party.
remove_linux_acls() {
    if [ "$OS" != "Linux" ]; then
        return
    fi
    if [ -x /usr/bin/setfacl ]; then
        printf "Found setfacl, removing ACLs for the install hierarchy...\n"
        setfacl --recursive --remove-all "${INSTALL_DIR:?}"
    fi
}

# Prevents using inaccessible paths such as /home/johndoe/Downloads/sftpplus.
validate_install_dir() {
    printf "Checking if \"%s\" user can access \"%s\"...\n" \
        "$SFTPPLUS_USER" "$INSTALL_DIR"
    # Do not chown the entire hierarchy yet, as it can take a while.
    chown "$SFTPPLUS_USER" "$INSTALL_INFO"
    if [ "$OS" = "Darwin" ]; then
        sudo -u "$SFTPPLUS_USER" touch "$INSTALL_INFO"
    else
        # Has to work everywhere else, don't use sudo or GNU-style options.
        su "$SFTPPLUS_USER" "-c touch $INSTALL_INFO"
    fi
}

configure_permissions() {
    printf "Adjusting user/group file permissions... "
    chown -R root:"$OS_ADM_GROUP" "$INSTALL_DIR"
    # Do not quote $SUBDIRS_TO_CHOWN, it's used as a list here.
    for sub_dir in $SUBDIRS_TO_CHOWN; do
        chown -R "$SFTPPLUS_USER":"$SFTPPLUS_USER" "${INSTALL_DIR:?}"/"$sub_dir"
        chmod -R o-rwx "${INSTALL_DIR:?}"/"$sub_dir"
    done
    printf "Done.\n"
}

suggest_services() {
    printf "\nAfter a successful start, check the default SFTPPlus services.\n"
    printf "For example, the Web Manager and Web Client interfaces at:\n"
    printf "    https://%s:%s\n" "$LOCAL_IP" "$PORT_LM"
    printf "    https://%s:%s\n" "$LOCAL_IP" "$PORT_HTTPS"
}

# First parameter is the needed setcap binary with full path.
# Second parameter is the package to check for and eventually install.
# Remaining parameters are the command used to install the package.
install_setcap_or_exit() {
    setcap_cmd="$1"; shift
    os_pkg="$1"; shift

    if [ ! -x "$setcap_cmd" ]; then
        yes_or_default_not "The \"$os_pkg\" package is needed. Install it?"
        if [ "$YES_OR_NOT" = "yes" ]; then
            "$@" "$os_pkg"
        else
            printf "To configure this manually, follow %s.\n" "$LNX_PORT_DOCS"
            exit 0
        fi
    fi
}

install_setcap_if_possible() {
    case "$LNX_DISTRO" in
        "alpine")
            install_setcap_or_exit /usr/sbin/setcap libcap apk add
            ;;
        "rhel"|"centos"|"ol"|"almalinux"|"rocky"|"amzn"|"fedora")
            install_setcap_or_exit /usr/sbin/setcap libcap yum -y install
            ;;
        "ubuntu"|"ubuntu-core"|"debian"|"devuan"|"linuxmint"|"pop"|"elementary")
            install_setcap_or_exit /sbin/setcap libcap2-bin apt -y install
            ;;
        *)
            # Unsupported distros reach this. Most should have setcap anyway.
            printf "Unknown package management."
            printf " Assuming libcap binary tools are present...\n"
            ;;
    esac
}

# Set extended attributes required by the Python executable
# to run SFTPPlus as instructed during installation.
set_capabilities() {
    if [ "$OS" != "Linux" ]; then
        printf "Write us at support@proatria.com to help you configure this.\n"
        return
    fi

    install_setcap_if_possible
    setcap 'cap_net_bind_service=+ep' "$INSTALL_DIR"/bin/python
    printf "Success! SFTPPlus has been configured to bind privileged ports.\n"
    printf "You may use ports below 1024 for SFTPPlus services from now on.\n"
}

ask_to_set_capabilities() {
    yes_or_default_not \
        "Do you want SFTPPlus to use ports below 1024, e.g. 21, 443?"
    if [ "$YES_OR_NOT" = "yes" ]; then
        set_capabilities
    else
        printf "If you later reconsider, "
        if [ "$OS" = "Linux" ]; then
            printf "follow %s.\n" "$LNX_PORT_DOCS"
        else
            printf "write us at support@proatria.com to help configure this.\n"
        fi
    fi
}

restore_capabilities() {
    if [ "$OS" != "Linux" ]; then
        return
    fi

    printf "Checking file capabilities to restore...\n"
    install_setcap_if_possible
    OLD_CAPS="$(getcap "$UPD_BKP_DIR"/bin/python)"
    if [ -z "$OLD_CAPS" ]; then
        printf "No capabilities found.\n"
        return
    fi

    # Check for an empty capability set which is not the same as no capability.
    CAPS_RAW_LIST="$(printf "%s" "$OLD_CAPS" | cut -d'=' -f2-)"
    if [ -z "$CAPS_RAW_LIST" ]; then
        setcap 'cap_net_bind_service=' "$INSTALL_DIR"/bin/python
        printf "Empty capability set restored.\n"
        return
    fi

    # The path to the old binary is in our backup path, so it has no spaces.
    CAPS_LIST="$(printf "%s" "$OLD_CAPS" | cut -d' ' -f2-)"
    setcap "$CAPS_LIST" "$INSTALL_DIR"/bin/python
    printf "Restored: %s\n" "$CAPS_LIST"
}

show_howto_start_manually() {
    printf "Alternatively, you may start SFTPPlus manually with:\n"
    printf "    su -l %s\n" "$DAEMON_USER"
    printf "    ./bin/admin-commands.sh start\n"
}

advise_container_autostart_and_exit() {
    printf "Container detected, not starting SFTPPlus automatically.\n"
    show_howto_start_manually
    exit 0
}

advise_unsupported_autostart() {
    (>&2 printf "Automatically starting SFTPPlus is ")
    (>&2 printf "not supported on this system.\n")
    printf "Please follow %s.\n" "$LNX_ADV_DOCS"
}

advise_manual_uninstall() {
    yes_or_default_not "Do you want to continue anyway?"
    if [ "$YES_OR_NOT" = "yes" ]; then
        return
    else
        printf "To manually uninstall SFTPPlus, please follow %s.\n" \
            "$UNINST_DOCS"
        exit 201
    fi
}

set_running_at_boot() {
    printf "\nSetting SFTPPlus to start on boot...\n"
    case "$OS" in
        "AIX")
            cp -f "$INSTALL_DIR"/bin/sftpplus-mft.sysv.sh \
                "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE"
            chmod +x "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE"
            ln -sf "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE" \
                /etc/rc.d/rc2.d/S99"$SFTPPLUS_SERVICE"
            ;;
        "Darwin")
            cp "$INSTALL_DIR"/bin/sftpplus-mft.plist \
                "$LAUNCHD_INIT_DIR"/"$SFTPPLUS_SERVICE".plist
            ;;
        "Linux")
            # Check for Docker container.
            if [ -f /.dockerenv ]; then
                # No need to add a capability under Docker.
                advise_container_autostart_and_exit
            fi
            if [ -d /run/systemd/system/ ]; then
                cp "$INSTALL_DIR"/bin/sftpplus-mft.service \
                    "$SYSTEMD_INIT_DIR"/"$SFTPPLUS_SERVICE".service
                systemctl enable "$SFTPPLUS_SERVICE".service
            elif [ "$LNX_DISTRO" = "alpine" ]; then
                cp "$INSTALL_DIR"/bin/sftpplus-mft.openrc.sh \
                    "$OPENRC_INIT_DIR"/"$SFTPPLUS_SERVICE"
                chmod +x "$OPENRC_INIT_DIR"/"$SFTPPLUS_SERVICE"
                rc-update add "$SFTPPLUS_SERVICE"
            else
                advise_unsupported_autostart
            fi
            ;;
    esac
}

uninstall_init_files() {
    printf "Disabling running SFTPPlus on boot, if enabled...\n"

    case "$OS" in
        "AIX")
            rm -f "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE" \
                /etc/rc.d/rc2.d/S99"$SFTPPLUS_SERVICE"
            ;;
        "Darwin")
            rm -f "$LAUNCHD_INIT_DIR"/"$SFTPPLUS_SERVICE".plist
            ;;
        "Linux")
            # Check for Docker container.
            if [ -f /.dockerenv ]; then
                printf "Container detected, not handling running on boot.\n"
                return
            fi
            if [ -d /run/systemd/system/ ]; then
                systemd_dir="$SYSTEMD_INIT_DIR"/multi-user.target.wants/
                if [ -f "$systemd_dir"/"$SFTPPLUS_SERVICE".service ]; then
                    systemctl disable "$SFTPPLUS_SERVICE".service
                fi
                rm -f "$SYSTEMD_INIT_DIR"/"$SFTPPLUS_SERVICE".service
            elif [ "$LNX_DISTRO" = "alpine" ]; then
                rm -f /etc/runlevels/*/"$SFTPPLUS_SERVICE" \
                    "$OPENRC_INIT_DIR"/"$SFTPPLUS_SERVICE"
            else
                (>&2 printf "\nDisabling SFTPPlus at boot is not ")
                (>&2 printf "supported on this Linux distribution.")
                advise_manual_uninstall
            fi
            ;;
    esac
}

get_linux_local_ip() {
    if [ -x /sbin/ip ]; then
        LOCAL_IP=$(/sbin/ip route get 1 | awk '{print $7}')
    # Next tests aren't united w/ -o as not all POSIX shells support it.
    elif [ -x /bin/hostname ]; then
        # Don't use -I, as BusyBox ersatz version doesn't support it.
        LOCAL_IP=$(hostname -i)
    elif [ -x /usr/bin/hostname ]; then
        # Don't use -I, as BusyBox ersatz version doesn't support it.
        LOCAL_IP=$(hostname -i)
    else
        LOCAL_IP="127.0.0.1"
    fi
}

# Used to stop the main SFTPPlus process, sftpplus-service-supervisor,
# when there is no option to use a service / init / unit file.
# Errors are expected if trying repeatedly in the recursing stop_if_running().
kill_server_supervisor() {
    set +o errexit
    if [ ! -s "$PID_FILE" ]; then
        (>&2 printf "SFTPPlus PID file is not present: %s\n" "$PID_FILE")
        printf "Stopping the supervisor SFTPPlus process owned by %s...\n" \
            "$DAEMON_USER"
        # pkill from procps is not good here, only matches up to 15 characters.
        # killall from Busybox not good enough. Add 'psmisc' package.
        killall -u "$DAEMON_USER" sftpplus-service-supervisor
        return
    fi
    server_pid="$(cat "$PID_FILE")"
    kill "$server_pid"
    set -o errexit
}

# FIXME:6540:
# Enable macOS ARM64 installer tests.
# This has only been superficially tested on GitHub's GHA instances.
kill_server_supervisor_macos() {
    sleep 3
    get_user_procs
    set +o errexit
    if [ -n "$USER_PROCS" ]; then
        pkill -u "$DAEMON_USER" -x sftpplus-service-supervisor
        kill_server_supervisor_macos
    fi
    set -o errexit
}

# On macOS, unloading through launchd doesn't stop the server properly.
macos_server_stop() {
    launchctl unload "$LAUNCHD_INIT_DIR"/"$SFTPPLUS_SERVICE".plist
    kill_server_supervisor_macos
}

# Detect best commands to start/stop the OS service
get_start_stop_cmds() {
    # Defaults. Also usable when install fails before copying init files.
    START_CMD=advise_unsupported_autostart
    STOP_CMD=kill_server_supervisor

    # Instead of defining generic commands per OS, actually check if the init
    # files were installed, otherwise it's better to stick to the defaults.
    if [ -f /.dockerenv ]; then
        # If running under Docker, there's no advanced init management.
        # Also exit there to avoid printing a no longer needed "Done".
        START_CMD=advise_container_autostart_and_exit
        STOP_CMD=kill_server_supervisor
    elif [ -f "$SYSTEMD_INIT_DIR"/"$SFTPPLUS_SERVICE".service ]; then
        START_CMD="systemctl start $SFTPPLUS_SERVICE.service"
        STOP_CMD="systemctl stop $SFTPPLUS_SERVICE.service"
    elif [ -f "$OPENRC_INIT_DIR"/"$SFTPPLUS_SERVICE" ]; then
        # Non-containerised Alpine or unsupported OpenRC/SysV system.
        START_CMD="service $SFTPPLUS_SERVICE start"
        STOP_CMD="service $SFTPPLUS_SERVICE stop"
    elif [ -f "$LAUNCHD_INIT_DIR"/"$SFTPPLUS_SERVICE".plist ]; then
        START_CMD="launchctl load $LAUNCHD_INIT_DIR/$SFTPPLUS_SERVICE.plist"
        STOP_CMD=macos_server_stop
    elif [ -f "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE" ]; then
        START_CMD="$AIX_INIT_DIR/$SFTPPLUS_SERVICE start"
        STOP_CMD="$AIX_INIT_DIR/$SFTPPLUS_SERVICE stop"
    fi
}

# The Busybox "applets" are not always usable.
alpine_exit_if_no_pkg() {
    apk_pkg="$1"

    if ! apk info -q -e "$apk_pkg"; then
        (>&2 printf "\nSFTPPlus service management requires ")
        (>&2 printf "the \"%s\" package on Alpine Linux.\n" "$apk_pkg")
        exit 190
    fi
}

# Only check for processes running as dedicated user.
# This is not perfect, but it's a necessity when uninstalling,
# as removing a user only works if there are no processes owned by it.
# Make sure these commands work for usernames longer than 8 characters.
get_user_procs() {
    # Commands below are allowed to fail as there may be no match.
    set +o errexit
    case "$OS" in
        "AIX")
            # shellcheck disable=SC2009 # No pgrep on AIX.
            USER_PROCS="$(ps aux | grep ^"$SFTPPLUS_USER")"
            ;;
        "Darwin")
            USER_PROCS="$(pgrep -lU "$SFTPPLUS_USER")"
            ;;
        "Linux")
            # On other distributions using Busybox, do the same manually.
            if [ "$LNX_DISTRO" = "alpine" ]; then
                # Busybox' pgrep shortens usernames longer than 8 characters.
                alpine_exit_if_no_pkg procps
                # Busybox' killall has no "-u" switch.
                # Although killall might never be needed, add it preventively.
                alpine_exit_if_no_pkg psmisc
            fi
            USER_PROCS="$(pgrep -lu "$SFTPPLUS_USER")"
            ;;
    esac
    set -o errexit
}

three_seconds_wait() {
    for _counter in 1 2 3; do
        printf "."
        sleep 1
    done
    printf "\n"
}

# This function recurses until there's no process running as SFTPPlus OS user.
stop_if_running() {
    if [ -z "$SFTPPLUS_USER" ]; then
        (>&2 printf "\nUnknown or unset system account. ")
        (>&2 printf "Running SFTPPlus services will not be stopped!")
        advise_manual_uninstall
        return
    fi

    get_user_procs

    if [ -z "$USER_PROCS" ]; then
        printf "SFTPPlus not currently running as user %s.\n" "$SFTPPLUS_USER"
        return
    fi

    printf "\nFound processes started as user %s:\n" "$SFTPPLUS_USER"
    printf "%s\n" "$USER_PROCS"

    # Avoid having the output above intertwined with the prompt below.
    sleep 1
    # Both $SFTPPLUS_SERVICE & $SFTPPLUS_USER are used below to give all hints.
    yes_or_default_not \
        "Automatically stop $SFTPPLUS_SERVICE service to continue?"
    if [ "$YES_OR_NOT" = "no" ]; then
        printf \
            "\nStop SFTPPlus services started as user %s before retrying.\n" \
            "$SFTPPLUS_USER"
        exit 200
    fi

    # Before stopping the running service, check if the system service exists.
    check_system_service_exists
    $STOP_CMD

    # Stopping takes some time, try avoiding racing issues.
    printf "Waiting for SFTPPlus services running as user %s to stop" \
        "$SFTPPLUS_USER"
    three_seconds_wait
    PROCESS_STOP_COUNTER=$((PROCESS_STOP_COUNTER+1))

    if [ "$PROCESS_STOP_COUNTER" -ge 20 ]; then
        (>&2 printf "\nCould not stop SFTPPlus after 20 retries...\n")
        (>&2 printf "Maybe something else restarts it independently?\n")
        exit 203
    fi

    stop_if_running
}

move_install_info() {
    old_install_info="${INSTALL_DIR:-}/lib/INSTALL_INFO"
    if [ -f "$old_install_info" ]; then
        printf "INSTALL_INFO found in deprecated location, moving it...\n"
        mv "$old_install_info" "$INSTALL_INFO"
    fi
}


# Exit with an error code if the saved SFTPPlus service
# from INSTALL_INFO doesn't match an existing OS service name.
exit_services_mismatch() {
    (>&2 printf "Could not find the system service: %s\n" "$SFTPPLUS_SERVICE")
    (>&2 printf "This indicates a mismatch between the enabled system services")
    (>&2 printf "\nand the SFTPPlus system service name saved at:")
    (>&2 printf "\n%s\n" "$INSTALL_INFO")
    printf "Before retrying, make sure the above service names match.\n"
    exit 202
}

check_system_service_exists() {
    if [ -f /.dockerenv ]; then
        # If running under Docker, there's no advanced init management.
        return
    fi

    case "$OS" in
        "AIX")
            if [ -f "$AIX_INIT_DIR"/"$SFTPPLUS_SERVICE" ]; then
                # Service file found, all good.
                return
            else
                exit_services_mismatch
            fi
            ;;
        "Darwin")
            if [ -f "$LAUNCHD_INIT_DIR"/"$SFTPPLUS_SERVICE".plist ]; then
                # Matching system daemon configuration file found, all good.
                return
            else
                exit_services_mismatch
            fi
            ;;
        "Linux")
            if [ -d /run/systemd/system/ ]; then
                # This is a systemd-based Linux distribution.
                if [ -f "$SYSTEMD_INIT_DIR"/"$SFTPPLUS_SERVICE".service ]; then
                    # Matching unit file found, all good.
                    return
                else
                    exit_services_mismatch
                fi
            fi

            if [ "$LNX_DISTRO" = "alpine" ]; then
                # This is an Alpine Linux system not running under Docker.
                if [ -f "$OPENRC_INIT_DIR"/"$SFTPPLUS_SERVICE" ]; then
                    # Matching OpenRC file found, all good.
                    return
                else
                    exit_services_mismatch
                fi
            fi
            ;;
    esac

    printf "\nCould not check for matching system service names. "
    printf "Continuing anyway...\n"
}

get_install_info() {
    move_install_info

    INSTALL_INFO_FOUND="no"
    if [ -f "$INSTALL_INFO" ]; then
        # Do not use "source" here, it's not working with Dash shell.
        # shellcheck source=/dev/null # The file definitely exists.
        . "$INSTALL_INFO"
        INSTALL_INFO_FOUND="yes"
    fi

    # These dirs are specific to the installed SFTPPlus instance.
    common_bkp_path_inc="$BKP_COMMON_DIR/$SFTPPLUS_SERVICE"
    UPD_BKP_DIR="${common_bkp_path_inc}_UPDATE_${BKP_SUFFIX}"
    ROLL_BKP_DIR="${common_bkp_path_inc}_ROLLBACK_${BKP_SUFFIX}"
    # These two dir names are used for creating the default links.
    UPD_BKP_GENERIC_LNK="${common_bkp_path_inc}_UPDATE_${BKP_STRING}"
    ROLL_BKP_GENERIC_LNK="${common_bkp_path_inc}_ROLLBACK_${BKP_STRING}"
}

get_os_user() {
    if [ "$INSTALL_INFO_FOUND" != "yes" ]; then
        (>&2 printf "Could not detect the OS user for the SFTPPlus process.\n")
        printf "Input operating system account used by SFTPPlus (sftpplus): "
        read_with_default SFTPPLUS_USER sftpplus
    fi
}

# FIXME:5956:
# Move updater version checks to Python.
# Check if the argument is a string with characters not allowed in a version.
check_version_string() {
    # Development versions also contain a dev-specific non-numeric string.
    ver_string="$(printf "%s" "$1" | sed s/"$DEV_STRING"//g)"
    # The initial x guards against an empty parameter.
    # The regex matches dots and numbers.
    if [ "$(expr "x$ver_string" : "x[.0-9]*$")" -eq 0 ]; then
        printf "Detected version \"%s\" contains unexpected characters.\n" "$1"
        exit 17
    fi
}

exit_if_trial() {
    if [ "${NEW_VER#"trial"}" != "$NEW_VER" ]; then
        printf "Updating to a trial version is not supported.\n"
        exit 0
    fi
}

exit_if_same_version() {
    if [ "$VERSION" != "$NEW_VER" ]; then
        return
    else
        printf "SFTPPlus version %s is already installed.\n" "$NEW_VER"
    fi

    # Things to do when compared versions are identical.
    if [ $# -eq 1 ]; then
        # If there's only one parameter, it's an online check.
        # Exit with non-zero error code, as the online version is not newer.
        exit 14
    elif [ $# -eq 2 ]; then
        # With only two parameters, it means there's no --force.
        # Exit with non-zero error code, as the update won't proceed.
        (>&2 printf "Append --force to proceed anyway.\n")
        exit 15
    elif [ "$3" = "--force" ]; then
        printf "Forcing the update...\n"
    fi
}

exit_if_not_forcing() {
    # The third parameter can only be "--force".
    if [ $# -lt 3 ]; then
        # The update was not forced.
        (>&2 printf "Downgrading is not supported.\n")
        if [ $# -eq 2 ]; then
            # There are two parameters, this is not an online check.
            # Advise on updating to provided package anyway.
            (>&2 printf "Append --force to proceed anyway.\n")
        fi
        exit 21
    fi

    if [ "$3" = "--force" ]; then
        printf "Forcing the update...\n"
    fi
}

# FIXME:5956:
# Move updater version checks to Python.
exit_if_no_update() {
    # Strip "+trial" from trial package version string.
    OLD_VER="${VERSION%+trial}"
    # The number of dots is exactly 2 in the version strings.
    for section_counter in 1 2 3; do
        # The sections are the portions of the version string delimited by dots.
        old_section="$(printf "%s" "$OLD_VER" | cut -d . -f $section_counter)"
        new_section="$(printf "%s" "$NEW_VER" | cut -d . -f $section_counter)"
        if [ "$old_section" -lt "$new_section" ]; then
            # Update detected, no need to check further.
            break
        elif [ "$old_section" -gt "$new_section" ]; then
            # Downgrade detected, exit if not forcing, and don't check further.
            (>&2 printf "Version %s is older. " "$NEW_VER")
            exit_if_not_forcing "$@"
            break
        fi
    done
}

# No package provided as parameter, but there's an update released online.
advise_to_update() {
    printf "Updating to version %s is possible using this script.\n" \
        "$NEW_VER"
    printf "\nTo proceed, manually download the SFTPPlus package\n"
    printf "for your operating system, then copy it to this server.\n"
    printf "Call this script again with the path to the downloaded "
    printf "file as a parameter.\n"
    printf "\nIf you prefer to update manually, please follow %s.\n" \
        "$UPGRADE_DOCS"
    # Only exit with 0 if there is an update available.
    exit 0
}

set_get_cmd() {
    GET_CMD="curl --silent"
    if [ "$OS" != "Linux" ]; then
        return
    fi
    if [ "$LNX_DISTRO" = "alpine" ]; then
        # Alpine's Busybox has a wget ersatz version.
        GET_CMD="wget -q -O -"
    fi
}

check_if_updating() {
    if [ $# -eq 1 ]; then
        # If no package provided, check latest version & advise accordingly.
        set_get_cmd
        # curl can error out silently because of cert issues on old systems.
        if ! $GET_CMD $LATEST_VER_LNK >/dev/null; then
            (>&2 printf "Can't check the latest version released online.\n")
            printf "Supply the path to a SFTPPlus package to update to.\n"
            exit 19
        fi
        NEW_VER="$($GET_CMD $LATEST_VER_LNK)"
        printf "Latest version released online: %s\n" "$NEW_VER"
        exit_if_same_version "$@"
        exit_if_no_update "$@"
        advise_to_update
    fi

    # There is a second parameter, check if it's a file.
    if [ ! -f "$2" ]; then
        (>&2 printf "Not a file: %s\n" "$2")
        printf "Supply the path to a SFTPPlus package to update to.\n"
        exit 20
    fi

    # The second parameter is a file, assumingly an SFTPPlus package.
    PKG_PATH="$2"
    PKG_NAME="$(basename "$PKG_PATH")"
    # Get the name of the unpacked sub-dir from the filename.
    # TAR_DIR is also used when moving the unpacked sub-dir.
    # This could break if the package filename is tampered with.
    case "$PKG_NAME" in
        *"$DEV_STRING"*)
            # Special case for development packages.
            TAR_DIR=$(printf "%s" "$PKG_NAME" | cut -d. -f1-4)
            ;;
        *)
            TAR_DIR=$(printf "%s" "$PKG_NAME" | cut -d. -f1-3)
            ;;
    esac
    NEW_VER=$(printf "%s" "$TAR_DIR" | cut -d- -f4)
    # For assurance, check if the version string matches requirements.
    check_version_string "$NEW_VER"
    printf "Provided package version found: %s\n" "$NEW_VER"

    exit_if_trial
    exit_if_same_version "$@"
    exit_if_no_update "$@"
}

# With -p, mkdir behaves properly everywhere anyway, but
# this saves an invocation of mkdir if the sub-dir already exists.
create_bkp_common_dir() {
    if [ ! -d "$BKP_COMMON_DIR" ]; then
        mkdir -p "$BKP_COMMON_DIR"
    fi
}

start_sftpplus_and_exit() {
    if [ "$SELINUX_ENFORCED" = "yes" ]; then
        restorecon "$ADMIN_CMD"
    fi
    printf "\nStarting SFTPPlus after %s...\n" "$ACTION"
    $START_CMD
    # systemd is silent when starting successfully.
    if [ -d /run/systemd/system/ ]; then
        printf "Done.\n"
    fi
    if [ "$ACTION" = "install" ]; then
        suggest_services
    fi
}

exit_if_bkp_dir_exists() {
    if [ -d "$1" ]; then
        (>&2 printf "\nHowever, the backup directory already exists:\n")
        (>&2 printf "%s\n" "$1")
        exit 210
    fi
}

update_bkp_dir_setup() {
    # Since version 4.30.0, this should a symbolic link.
    if [ ! -L "$UPD_BKP_GENERIC_LNK" ]; then
        # However, previous versions had that as the default backup sub-dir.
        if [ -d "$UPD_BKP_GENERIC_LNK" ]; then
            printf "\nFound old generic backup SFTPPlus sub-directory at:\n"
            printf "%s" "$UPD_BKP_GENERIC_LNK"
            printf "\nMoving it to:\n"
            printf "%s.old...\n\n" "$UPD_BKP_GENERIC_LNK"
            mv "$UPD_BKP_GENERIC_LNK" "$UPD_BKP_GENERIC_LNK".old
        fi
    fi
}

update_sftpplus() {
    # Start with unpacking, as it can take some time and even error out.
    printf "Unpacking to /tmp the installation package \"%s\"...\n" "$PKG_NAME"
    # Portable invocations of both commands.
    gunzip -cd "$PKG_PATH" | tar -xf - -C /tmp/

    create_bkp_common_dir
    update_bkp_dir_setup
    stop_if_running
    printf "\nMoving existing SFTPPlus installation to:\n"
    printf "%s\n" "$UPD_BKP_DIR"
    mv "$INSTALL_DIR" "$UPD_BKP_DIR"

    printf "Updating generic backup SFTPPlus link.."
    rm -f "$UPD_BKP_GENERIC_LNK"
    ln -s "$UPD_BKP_DIR" "$UPD_BKP_GENERIC_LNK"
    printf ". Done.\n"

    printf "Moving unpacked SFTPPlus files from /tmp to \"%s\"...\n" \
        "$INSTALL_DIR"
    mv /tmp/"$TAR_DIR" "$INSTALL_DIR"

    # Permissions are restored immediately after unpacking.
    # For files copied from the backup, permissions are preserved through "cp".
    get_os_user
    configure_permissions
    restore_capabilities

    # The new files in "extension/" are preserved to overwrite the old ones.
    mv "$INSTALL_DIR"/extension "$INSTALL_DIR"/extension.new
    # These sub-dirs should be empty in the installation tarball.
    rmdir "$INSTALL_DIR"/configuration "$INSTALL_DIR"/log

    printf "\nRestoring from previous installation:"
    # Counter used for checking if backed-up stuff is to be restored or not.
    total_subdirs_to_remove=$(printf "%s" "$SUBDIRS_TO_REMOVE" | wc -w)
    # Elements in the backed-up dir can be sub-dirs or regular files.
    for element in "$UPD_BKP_DIR"/*; do
        counter=0
        for sub_dir in $SUBDIRS_TO_REMOVE; do
            if [ "${element%/"$sub_dir"}" != "$element" ]; then
                # The element matches a sub-dir not to be restored, like "bin".
                break
            fi
            counter=$((counter+1))
        done
        # Counter is reached when the current element doesn't match any sub-dir
        # not to be restored, therefore it has to be restored from the backup.
        if [ "$counter" -eq "$total_subdirs_to_remove" ]; then
            # The element can be a sub-dir like configuration/, extension/,
            # log/, or other custom files and dirs from the backed-up directory.
            printf " %s" "${element#"$UPD_BKP_DIR"/}"
            cp -pR "$element" "$INSTALL_DIR"/
        fi
    done
    printf ". Done.\n"

    # Files in extension.new/ were saved to overwrite old ones with same name.
    printf "Copying over the new packaged files in: extension"
    mv "$INSTALL_DIR"/extension.new/* "$INSTALL_DIR"/extension/
    rmdir "$INSTALL_DIR"/extension.new
    printf ". Done.\n"

    if [ -f "$POST_UPDATE_SCRIPT" ]; then
        printf "\nFound post-install script, executing it...\n"
        # Launching it this way, only exported vars are kept, e.g. $INSTALL_DIR.
        "$POST_UPDATE_SCRIPT"
    fi

    printf "\nSuccessfully updated SFTPPlus at \"%s\".\n" "$INSTALL_DIR"
    printf "Previous version is backed up at:\n"
    printf "%s\n" "$UPD_BKP_DIR"
    start_sftpplus_and_exit
}

rollback_sftpplus(){
    create_bkp_common_dir
    printf "\nMoving existing SFTPPlus installation to:\n"
    printf "%s\n" "$ROLL_BKP_DIR"
    mv "$INSTALL_DIR" "$ROLL_BKP_DIR"

    printf "Updating generic rolled-back SFTPPlus link.."
    rm -f "$ROLL_BKP_GENERIC_LNK"
    ln -s "$ROLL_BKP_DIR" "$ROLL_BKP_GENERIC_LNK"
    printf ". Done.\n"

    printf "Copying backup SFTPPlus installation from:\n"
    printf "%s\n" "$UPD_BKP_GENERIC_LNK"
    # Permissions are preserved with "-p".
    # Symbolic links are always followed.
    cp -pRL "$UPD_BKP_GENERIC_LNK" "$INSTALL_DIR"

    printf "Copying over files from rolled-back installation: log"
    rm -rf "${INSTALL_DIR:?}"/log
    cp -pR "$ROLL_BKP_DIR"/log "$INSTALL_DIR"/
    printf ". Done.\n"

    printf "\nSuccessfully rolled back SFTPPlus at \"%s\".\n" "$INSTALL_DIR"
    printf "Previously-installed version is backed up at:\n"
    printf "%s\n" "$ROLL_BKP_DIR"
    start_sftpplus_and_exit
}

uninstall_sftpplus() {
    uninstall_init_files

    printf "Removing SFTPPlus sub-directories:"
    # Do not quote $SUBDIRS_TO_REMOVE, it's used as a list here.
    for sub_dir in $SUBDIRS_TO_REMOVE; do
        if [ -d "${INSTALL_DIR:?}"/"$sub_dir" ]; then
            printf " %s/" "$sub_dir"
            rm -rf "${INSTALL_DIR:?}"/"$sub_dir"
        fi
    done
    printf "\nAll the other files in the SFTPPlus install path were kept.\n"

    if [ "$INSTALL_INFO_FOUND" = "yes" ]; then
        printf "Removing \"%s\" user/group...\n" "$SFTPPLUS_USER"
        case "$OS" in
            "AIX")
                userdel "$SFTPPLUS_USER"
                rmgroup "$SFTPPLUS_USER"
                ;;
            "Darwin")
                dscl . delete /Users/"$SFTPPLUS_USER"
                dscl . delete /Groups/"$SFTPPLUS_USER"
                ;;
            "Linux")
                if [ "$LNX_DISTRO" = "alpine" ]; then
                    deluser "$SFTPPLUS_USER"
                else
                    userdel "$SFTPPLUS_USER"
                    if [ "$SELINUX_ENFORCED" = "yes" ]; then
                        printf "Removing /var/lib/%s...\n" "$SFTPPLUS_USER"
                        rmdir /var/lib/"$SFTPPLUS_USER"
                    fi
                fi
                ;;
        esac
    else
        printf "\nPreviously-saved installation details were not found at\n"
        printf "\"%s\".\n" "$INSTALL_INFO"
        printf "SFTPPlus system user/group were not removed.\n"
    fi

    printf "\nSuccessfully uninstalled SFTPPlus.\n"
    printf "Existing configuration and log files were preserved at \"%s\".\n" \
        "$INSTALL_DIR"
}


#
# The script starts here.
#

if [ $# -lt 1 ]; then
    (>&2 printf "Unsupported invocation of this helper script.\n")
    exit 89
fi
# The variable is also used for some generic messages and prompts.
ACTION="$1"

# This is how errors are caught to insert an informative message at the end.
# The non-standard ERR signal is not supported on some POSIX shells like Dash.
# Therefore, trap EXIT and check the exit code instead.
# Doesn't work with the default /bin/sh on AIX, but doesn't break things either.
trap 'error_out_with_generic_message' EXIT

# This checks if $BKP_STRING is part of the $INSTALL_DIR string.
case "$INSTALL_DIR" in
    *"$BKP_STRING"*)
        (>&2 printf "You are trying to %s SFTPPlus at:\n" "$ACTION")
        (>&2 printf "%s.\n" "$INSTALL_DIR")
        (>&2 printf "That path is detected as an SFTPPlus backup directory.\n")
        printf "\n(This might be caused by the shell residing in "
        printf "the initial installation path.\n"
        printf "If that is the case, change directory to any path, "
        printf "including \".\", then retry.)\n"
        exit 88
        ;;
esac

OS=$(uname -s)
case "$OS" in
    "AIX")
        OS_ADM_GROUP="adm"
        LOCAL_IP=$(lsattr -El en0 | grep ^'netaddr ' | awk '{print $2}')
        export PATH="$PATH":/opt/freeware/bin
        ;;
    "Darwin")
        OS_ADM_GROUP="admin"
        LOCAL_IP=$(osascript -e "IPv4 address of (system info)")
        ;;
    "Linux")
        OS_ADM_GROUP="root"
        if [ -f /usr/sbin/getenforce ]; then
            if [ "$(/usr/sbin/getenforce)" = "Enforcing" ]; then
                export SELINUX_ENFORCED="yes"
            fi
        fi
        if [ ! -f /etc/os-release ]; then
            # Some old supported distributions such as RHEL 6 don't have it.
            (>&2 printf "Could not determine current Linux distribution.\n")
            printf "If using GLIBC 2.%s or newer, " "$MIN_GLIBC2_VER"
            printf "please set up SFTPPlus manually\n"
            printf "following %s.\n" "$LNX_DOCS"
            exit 91
        else
            # Do not use "source" here, it's not working with Dash shell.
            # shellcheck source=/dev/null # The file definitely exists.
            . /etc/os-release
            # Export it to have it in the post-update.sh script as well.
            export LNX_DISTRO="$ID"
            get_linux_local_ip
        fi
        ;;
    *)
        printf "Unsupported OS: %s.\n" "$OS"
        exit 99
        ;;
esac

if [ "$(id -u)" -ne 0 ]; then
    (>&2 printf "This script requires elevated privileges. ")
    (>&2 printf "Retry as root or using sudo.\n")
    exit 100
fi

case "$ACTION" in
    "install")
        if [ -f "$INSTALL_CONF" ]; then
            (>&2 printf "SFTPPlus configuration already exists. ")
            (>&2 printf "Initialization aborted.\n")
            exit 101
        fi
        printf "Initiating SFTPPlus install in \"%s\"...\n" "$INSTALL_DIR"
        check_admin_cmd
        get_daemon_and_sftpplus_users
        initialize_config
        # Configure OS user after initializing config, to properly chown it.
        add_os_user
        remove_linux_acls
        validate_install_dir
        configure_permissions
        set_running_at_boot
        get_start_stop_cmds
        if [ "$DAEMON_USER" != "root" ]; then
            ask_to_set_capabilities
        else
            printf "\nSFTPPlus is set to start as root, you may configure it\n"
            printf "to use ports below 1024 for its services, e.g. 21, 443.\n"
        fi
        printf "\nSFTPPlus has been set up successfully.\n"
        start_sftpplus_and_exit
        ;;
    "update")
        printf "Initiating SFTPPlus update at \"%s\"...\n" "$INSTALL_DIR"
        check_admin_cmd
        get_install_info
        if [ ! -f "$INSTALL_DIR"/$CONF_FILE ]; then
            (>&2 printf "\n%s doesn't seem to hold a " "$INSTALL_DIR")
            (>&2 printf "configured SFTPPlus installation. Update aborted.\n")
            printf "Maybe you need to run %s/install.sh first?\n" "$INSTALL_DIR"
            exit 140
        fi
        printf "Currently installed version is: %s\n" "$VERSION"

        check_if_updating "$@"
        printf "\nAll files installed by SFTPPlus are to be updated,\n"
        printf "excluding init-related service files, if applicable.\n"
        printf "Logs and configuration files are persisted.\n"
        printf "The current installation is going to be backed up at:\n"
        printf "%s\n" "$UPD_BKP_DIR"
        exit_if_bkp_dir_exists "$UPD_BKP_DIR"
        get_start_stop_cmds
        yes_or_default_not "Start the update of SFTPPlus at \"$INSTALL_DIR\"?"
        if [ "$YES_OR_NOT" = "yes" ]; then
            update_sftpplus
        fi
        ;;
    "rollback")
        if [ $# -gt 1 ]; then
            printf "Custom directory provided for rolling back: %s\n" "$2"
            UPD_BKP_DIR="$2"
        fi
        get_install_info
        if [ ! -f "$UPD_BKP_GENERIC_LNK"/$CONF_FILE ]; then
            (>&2 printf "\nRollback aborted. No SFTPPlus backup found at:\n")
            (>&2 printf "%s\n" "$UPD_BKP_GENERIC_LNK")
            exit 151
        fi
        printf "Initiating SFTPPlus rollback from:\n"
        printf "%s\n" "$UPD_BKP_GENERIC_LNK"
        printf "\nAll files installed by SFTPPlus are to be rolled back,\n"
        printf "excluding init-related service files, if applicable.\n"
        printf "The log files of the current installation are preserved.\n"
        printf "Current version is going to be backed up at:\n"
        printf "%s\n" "$ROLL_BKP_DIR"
        exit_if_bkp_dir_exists "$ROLL_BKP_DIR"
        get_start_stop_cmds
        stop_if_running
        yes_or_default_not "Rollback SFTPPlus at \"$INSTALL_DIR\"?"
        if [ "$YES_OR_NOT" = "yes" ]; then
            rollback_sftpplus
        fi
        ;;
    "uninstall")
        printf "Initiating SFTPPlus uninstall from \"%s\"...\n" "$INSTALL_DIR"
        printf "\nAll files installed by SFTPPlus are to be removed,\n"
        printf "including init-related service files, if applicable.\n"
        printf "Logs and configuration directories/files are not removed.\n"
        printf "System account is only removed if added by SFTPPlus install.\n"
        get_install_info
        get_start_stop_cmds
        stop_if_running
        yes_or_default_not "Really uninstall SFTPPlus from \"$INSTALL_DIR\"?"
        if [ "$YES_OR_NOT" = "yes" ]; then
            uninstall_sftpplus
        fi
        ;;
    *)
        (>&2 printf "Unsupported invocation of this helper script.\n")
        printf "Use the scripts in the \"bin/\" sub-directory of your "
        printf "SFTPPlus installation:\n"
        printf "install.sh, update.sh, rollback.sh, uninstall.sh.\n"
        exit 13
        ;;
esac
