#!/bin/sh
#
# Copyright 2022 ProAtria team
#
# Script for updating SFTPPlus on Linux / Unix.
# Uses the default /bin/sh shell. Tested with the OS'es supported by SFTPPlus.
#
# It supports updating to a newer version with the downloaded archive passed
# as a parameter, for example:
#
#     /opt/sftpplus/bin/update.sh /tmp/sftpplus-lnx-x64-4.24.0.tar.gz

if ! command -v dirname >/dev/null 2>&1; then
    printf "Missing \"dirname\" command.\n"
    exit 254
fi

# POSIX shell assurances.
set -o nounset
set -o errexit

bin_relative_path="$(dirname "$0")"

# Using pwd in a sub-shell, get the absolute path to the sub-dir of this script.
bin_path=$(cd "$bin_relative_path" && pwd)

if [ -z "$bin_path" ]; then
    printf "Couldn't determine absolute path for \"%s\".\n" "$0"
    exit 253
fi

export INSTALL_DIR="${bin_path%/bin}"

if [ "$(uname)" = "AIX" ]; then
    # Avoid the error "@: parameter not set" when there is no parameter,
    # as the behaviour of the default /bin/sh is different on AIX.
    set +o nounset
fi

# Show this help message and exit.
# Whenever editing this, check alignment of opts / associated text and length.
command_show_help() {
    printf "\nUsage:\n"
    printf "  %s [PATH_TO_TAR_GZ]\n" "$0"
    printf "\nOptions:\n"
    printf "  -h, --help\t\t\tShow this help message and exit\n"
    printf "  PATH_TO_TAR_GZ\t\tPath to an SFTPPlus installation package\n"
    printf "\nThe provided package, for example"
    printf " \"sftpplus-linux-x64-5.1.0.tar.gz\",\n"
    printf "is used to update the current SFTPPlus setup.\n"
    printf "\nWhen no installation package is provided, this script checks\n"
    printf "if there is a newer released package at https://www.sftpplus.com.\n"
    exit 0
}

# Get command line parameters and act accordingly.
while getopts "h-:" opt; do
    case $opt in
        h)
            command_show_help
            ;;
        -)
            # getopts doesn't support long opts such as --help.
            case "$OPTARG" in
                help)
                    command_show_help
                    ;;
                *)
                    printf "Invalid option: --%s\n" "$OPTARG"
                    exit 1
                    ;;
            esac
            ;;
        \?)
            # Unknown option.
            exit 1
            ;;
    esac
done

exec /bin/sh "$INSTALL_DIR"/bin/setup_helper.sh update "$@"
